#include <stdio.h>
#include <math.h>
#include "parms.h"
#include "graddata.h"
#include "gradient.h"
#include "bfgs.h"
#define INITIDENTITY 1
#define ITMAX 300     /* max number of small iterations */
#define MAXBIG 100    /* max number of big iterations */
#define EPS 0.000000000000001   /* machine precision */
#define TOLX (2000*EPS) /* convergence criterion on x values */
#define STPMX 100000000.0      /* scaled maximum step length allowed in 
                        line searches */

static char *sccsversion = "%W%   %G%";

short int stepcheck;
double newstep;

/*defs */
static double maxarg1, maxarg2;
#define FMAX(a,b) (maxarg1=(a), maxarg2=(b), (maxarg1) < (maxarg2) ?\
(maxarg1) : (maxarg2))

void dfgsmin(double *p, int n, double gtol, int *iter,
	     double *fret, double hessin[VARS][VARS], double **domains)
{
  extern double (*func4g)(double *);
  extern void (*func4boundarycheck)(double *);
  extern int useyear, usexfer, usepac;
  extern int *onboundary;

  extern short int stepcheck;

  int nb;

  int flag, gflag, bflag;
  int check, i, ii, its, j, bigloop;
  int nullity;
  double den, fac, fad, fae, fp, stpmax, sum=0.0,sumdg,sumxi, temp,test;
  double dg[VARS], g[VARS], hdg[VARS], pnew[VARS], xi[VARS];
  double pb[VARS], gb[VARS];

  double (*func)(double *) = func4g;

  func4boundarycheck(p-1);
  for (nb=0, i=0; i<n; i++) if (onboundary[i]==0) nb++;
  printf("onboundary: ");
  for (i=0; i<n; i++) printf("%d", onboundary[i]);
  printf("\n");

  fp=(double) func(p); /* calculate starting function value and gradient */
  gradient(p,g,n);  /* calculate the original gradient */

  /* copy subvectors for parms that are not on the boundary */
  for (ii=0, i=0; i<n; i++)
    if (onboundary[i]==0) {
      pb[ii] = p[i];
      gb[ii] = g[i];
      ii++;
    }

  newstep=STPMX;

  /*printf("Starting BFGS\n");*/
  stepcheck=1;
  for(bigloop=1;bigloop<=MAXBIG;bigloop++)  {

    /*
    if (stepcheck == 0) {
      printf("Bigloop exit\n");
      break;
    }
    */

    if (stepcheck == 1)
      stepcheck=0;

    ihessinit(hessin, nb, gb, pb, xi, &sum, INITIDENTITY);
    stpmax=newstep*FMAX(sqrt(sum),(double)nb);

    gflag = flag = 0;
    /* INNER LOOP OVER ITERATIONS */
    for (its=1;its<=ITMAX;its++) {
      *iter=its;

      /* note:  the lnsrch call uses p (off-boundary elements) as a work array */
      lnsrch(nb,pb,fp,gb,xi,pnew,fret,stpmax,&check,n,p);

      if (stepcheck == 1) {
	printf("STEP SIZE DECREASE: breaking inner loop\n");
	newstep=newstep/2;
	break;
      }

      fp= *fret;

      /* update line direction, and the current point */
      for (i=0;i<nb;i++) {
	xi[i]=pnew[i]-pb[i];
	pb[i]=pnew[i];
      }
      /* copy back parms that are not on the boundary */
      for (ii=0, i=0; i<n; i++)
	if (onboundary[i]==0) {
	  p[i] = pb[ii];
	  ii++;
	}

      /* test for convergence */
      test=0.0;
      for (i=0;i<nb;i++) {
	temp=fabs(xi[i])/FMAX(fabs(pb[i]),1.0);
	if (temp > test) test=temp;
      }

      if (its > (1<ITMAX?1:0) && test < TOLX) {
	printf("\nBIG(%i):%i\n",bigloop,its);
	printf("The current parms and gradients at fit value(%lf) are:\n",*fret);
	for(i=0;i<n;i++)
	  printf("X[%i] = %lf; \t grad= %lf\n",i+1, p[i], g[i]);
	fflush(stdout);
        if (flag==1) {
	  printf("TOLX exit\n");
	  return;
	}
	flag=1;
      }
      else
	flag = 0;

      /* Save the old Gradient, and get the new gradient.
	 Test for convergence on zero gradient */
      for (i=0;i<nb;i++) dg[i]=gb[i];

      gradient(p,g,n);

      /* copy subvectors for parms that are not on the boundary */
      for (ii=0, i=0; i<n; i++)
	if (onboundary[i]==0) {
	  gb[ii] = g[i];
	  ii++;
	}

      test=0.0;
      den=FMAX(*fret, 1.0);

      for (i=0;i<nb;i++)
	{
	  temp=fabs(gb[i])*FMAX(fabs(pb[i]),1.0)/den;
	  if (temp > test) test=temp;
	}

      if (its > (1<ITMAX?1:0) && test < gtol) {
	printf("\nBIG(%i):%i\n",bigloop,its);
	printf("The current parms and gradients at fit value(%lf) are:\n",*fret);
	for(i=0;i<n;i++)
	  printf("X[%i] = %lf; \t grad= %lf\n",i+1, p[i], g[i]);
	fflush(stdout);
	if (gflag==1) {
	  printf("GTOL exit\n");
	  return;
	}
	gflag=1;
      }
      else
	gflag = 0;

      /* Compute difference of gradients, and difference times current matrix */
      for (i=0;i<nb;i++) dg[i]=gb[i]-dg[i];
      for (i=0;i<nb;i++)
	{
	  hdg[i]=0.0;
	  for (j=0;j<nb;j++) hdg[i] += hessin[i][j]*dg[j];
	}

      /* Calculate dot products for the denomminators */
      fac=fae=sumdg=sumxi=0.0;
      for (i=0;i<nb;i++)
	{
	  fac += dg[i]*xi[i];
	  fae += dg[i]*hdg[i];
	  sumdg += dg[i]*dg[i];
	  sumxi += xi[i]*xi[i];
	}

      /* Skip update if fac not sufficiently positive */
      if (fac*fac > EPS*sumdg*sumxi)
	{
	  fac=1.0/fac;
	  fad=1.0/fae;
	  /* The Vector that makes BFGS different from DFP */
	  for (i=0;i<nb;i++) dg[i]=fac*xi[i]-fad*hdg[i];
	  /* The BFGS Update Formula */
	  for (i=0;i<nb;i++)
	    {
	      for (j=0;j<nb;j++)
		{
		  hessin[i][j] += fac*xi[i]*xi[j]-fad*hdg[i]*
		    hdg[j]+fae*dg[i]*dg[j];
		}
	    }
	}
      /* check that inverse hessian is positive definite */
      bfgscholesky (hessin, nb, &nullity);
      if (nullity != 0) {
	printf("bfgs:  singular hessin (%d), reverting to initial\n", nullity);
	ihessinit(hessin, nb, gb, pb, xi, &sum, INITIDENTITY);
	fflush(stdout);
      }

      /* The next direction to go in */
      for (i=0;i<nb;i++)
	{
	  xi[i]=0.0;
	  for (j=0;j<nb;j++) xi[i] -= hessin[i][j]*gb[j];
	}

    } /* end of iteration loop */
    printf("Reached limit for inner loop at BIG(%i):%i\n",bigloop,its);
    printf("Intermediate parms and gradients at fit value(%lf) are:\n",*fret);
    bflag = 0;
    for (ii=0, i=0; i<n; i++)
      if (onboundary[i]==0) {
	if ((pb[ii]<domains[i+1][1]) || (pb[ii]>domains[i+1][3])) {
	  bflag=i+1;
	  break;
	}
	ii++;
      }
    if (bflag==0)
      for(i=0;i<n;i++) {
	printf("X[%i] = %lf; \t with grad=%lf\n",i+1, p[i], g[i]);
      }
    else {
      printf("OOB at X[%d] (stopping dfgsmin)\n", bflag);
      break;
    }
    fflush(stdout);
  } /* end of bigloop */

  printf("The job ended at BIG(%i):%i\n",bigloop,its);
  printf("End of Loop\n");
  printf("The final parms and gradients at fit value(%lf) are:\n",*fret);
  for(i=0;i<n;i++) {
    printf("X[%i] = %lf; \t with grad=%lf\n",i+1, p[i], g[i]);
  }

} /* end of function */


/* THE LNSRCH CALL */
#define ALF 1.0e-4  /* ensures sufficient decrease in function */

void lnsrch(int n, double *xold, double fold, double *g, double *p,
            double *x, double *f, double stpmax, int *check,
	    int nall, double *pall)
{
  extern double (*func4g)(double *);
  extern int *onboundary;

  int i, ii;
  double a, alam, alam2, alamin, b, disc, f2, fold2, rhs1, rhs2, slope,
  sum, temp, test, tmplam;

  double (*func)(double *) = func4g;

  *check=0;
  for (sum=0.0,i=0;i<n;i++) sum += p[i]*p[i];
  sum=sqrt(sum);
  if (sum > stpmax)
    for (i=0; i<n; i++) p[i] *= stpmax/sum; /* scale if attempted step is too big */
  for (slope=0.0, i=0;i<n;i++)
    slope += g[i]*p[i];
  test=0.0;
  for (i=0;i<n;i++) {
    temp=fabs(p[i])/FMAX(fabs(xold[i]),1.0);
    if (temp > test) test=temp;
  }

  alamin=TOLX/test;
  alam=1.0;
  for (;;) {
    for (i=0;i<n;i++) x[i]=xold[i]+alam*p[i];

    /* printf("\n the x is:\n");
       fflush(stdout);
       for (i=0;i<n;i++) printf("X[%i]=%lf\n", i, x[i]); */

    /* copy parms into full-size vector */
    for (ii=0, i=0; i<nall; i++)
      if (onboundary[i]==0) {
	pall[i] = x[ii];
	ii++;
      }
    *f=(double) func(pall);

    /*  if (stepcheck == 1) {
	printf("STEP SIZE CHANGE: leaving lnsrch\n");
	return;
	} */

    if (alam < alamin) {
      for (i=0;i<n;i++) x[i]=xold[i];
      *check=1;
      return;
    } else if (*f <= fold+ALF*alam*slope) return;
    else {
      if (alam == 1.0)
	tmplam = -slope/(2.0*(*f-fold-slope));

      else {
	rhs1 = *f-fold-alam*slope;
	rhs2=f2-fold2-alam2*slope;
	a=(rhs1/(alam*alam)-rhs2/(alam2*alam2))/(alam-alam2);
	b=(-alam2*rhs1/(alam*alam)+alam*rhs2/(alam2*alam2))/(alam-alam2);
	if (a == 0.0) tmplam = -slope/(2.0*b);
	else {
	  disc=b*b-3.0*a*slope;
	  if (disc<0.0) {
	    printf("\nLNSRCH: stepcheck=1.\n");
	    stepcheck=1;
	    return;
	  }
	  else tmplam=(-b+sqrt(disc))/(3.0*a);
	}
	if (tmplam>0.5*alam) tmplam=0.5*alam;
      }
    }

    alam2=alam;
    f2= *f;
    fold2=fold;
    alam=FMAX(tmplam,0.1*alam);
  }

} /* end of procedure */

void ihessinit (double hessin[VARS][VARS], int nrow, double *g, double *p,
		double *xi, double *sum, int iflag)
{
  extern int useyear, usexfer, usepac;

  int i,j, nullity;
  double hessian[VARS][VARS];
  double tmp;

  *sum = 0.0;

  if (iflag==1) {
    for (i=0;i<nrow;i++) {
      for (j=0;j<nrow;j++)
	hessin[i][j]= 0.0;
      hessin[i][i]= 1.0;
      xi[i]=-g[i];
      *sum += p[i]*p[i];
    }
  }
  else {
    /* initialize the est. inverse Hessian from global array */
    for (i=0;i<nrow;i++) {
      xi[i]=0.0;
      for (j=0;j<nrow;j++) {
	hessin[i][j]= ihessians[(i*nrow)+j];
	xi[i] -= hessin[i][j]*g[j];
      }
    }
    /* check that inverse hessian is positive definite */
    bfgscholesky (hessin, nrow, &nullity);
    if (nullity == 0) {
      bfgsinverse(nrow, hessin, hessian);
      for (i=0;i<nrow;i++)
	for (j=0;j<nrow;j++)
	  *sum += p[i]*p[j]*hessian[i][j];  /* use hessian as metric */
    }
    else {
      printf("bfgs: singular initial hessin (%d), reverting to diagonal\n", nullity);
      fflush(stdout);
      /* code to initialize the est. inverse Hessian to diagonal matrix */
      for (i=0;i<nrow;i++) {
	for (j=0;j<nrow;j++)
	  hessin[i][j]= 0.0;
	hessin[i][i]= ihessians[(i*nrow)+i];
	if (hessin[i][i] <= 0.0) hessin[i][i] = 1.0;
	xi[i]=-g[i];
	*sum += p[i]*p[i]*hessin[i][i];
      }
    }
  }
  for (tmp=0.0, i=0;i<nrow;i++) tmp += p[i]*p[i];
  printf("ihessinit:  sum = %lf, two-norm = %lf\n", *sum, tmp);
}

#define TOOSMALL ((double) 2.0e-14)
void bfgscholesky (double hessin[VARS][VARS], int nrow, int *nullity)
{
  int i, j, k, row;
  double sum;
  double w[VARS][VARS];

  *nullity = 0;
  for (row=0; row<nrow; row++) {
    /* First compute w[row][row] */
    sum = hessin[row][row];
    for (j=0; j<row; j++) sum -= w[j][row]*w[j][row];
    if (sum > TOOSMALL) {
      w[row][row] = sqrt(sum);
      /* Now find elements w[row][k], k > row. */
      for (k=(row+1); k<nrow; k++) {
	sum = hessin[row][k];
	for (j=0; j<row; j++) 
	  sum -= w[j][row]*w[j][k];
	w[row][k] = sum/w[row][row];
      }
    }
    else {  /* blast off the entire row. */
      for (k=row; k<nrow; k++) w[row][k] = 0.0;
      (*nullity)++;
    }
  }
}

/* use the genoud inverse function to invert a matrix */
void bfgsinverse(int nrow, double matin[VARS][VARS], double matout[VARS][VARS])
{
  extern void inverse();
  extern double **matrix();

  static int firsttime = 0;
  static double **a, **a_inverse;

  int i,j;

  if (firsttime==0) {
    a = matrix(1, nrow, 1, nrow);
    a_inverse = matrix(1, nrow, 1, nrow);
  }

  for (i=0; i<nrow; i++)
    for (j=0; j<nrow; j++)
      a[i+1][j+1] = matin[i][j];

  inverse(a,a_inverse,nrow);

  for (i=0; i<nrow; i++)
    for (j=0; j<nrow; j++)
      matout[i][j] = a_inverse[i+1][j+1];
}
