#include "gendop.h"
#include "parms.h"
#include "gradient.h"
#include "bfgs.h"

static char *sccsversion = "%W%   %G%";

#if DOS_SYS
extern FILE *input,*output;
extern int test_num;
#endif

/* Cummulative probability on crossover */
/* Random probability on mutation       */
/* NO multiple hits per agent possible  */


/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   optimization()                               */
/*                                                                              */
/*           SYNOPSIS          :   void optimization(X,x1,x2,fin_mat,rc,tot_eq) */
/*                                                                              */
/*           DESCRIPTION       :   This procedure initializes the population    */
/*                                  with the values X passed from main, and     */
/*                                  evaluates them.  After assigning weight     */
/*                                  for each member of the populaiton, a group  */
/*                                  of them are chosen to reproduce and a group */
/*                                  is chosen to die.  Genetic operators are    */
/*                                  applied and a new generation is produced    */
/*                                  to replace the members that died.  This     */
/*                                  cycle continues for the number of times,    */
/*                                  user specifies in the input file            */
/*                                                                              */
/*           FUNCTIONS CALLED  :   assign_probab(),                             */
/*                                 evaluate(),                                  */
/*                                 find_cum_probab(),                           */
/*                                 find_live_die(),                             */
/*                                 find_parent(),                               */
/*                                 ivector(),                                   */
/*                                 matrix(),                                    */
/*                                 oper1(),                                     */
/*                                 oper2(),                                     */
/*                                 oper3(),                                     */
/*                                 oper4(),                                     */
/*                                 oper5(),                                     */
/*                                 oper6(),                                     */
/*                                 print_matrix(),                              */
/*                                 print_population(),                          */
/*                                 sort(),                                      */
/*                                 vector().                                    */
/*                                                                              */
/*           CALLING FUNCITONS :   main()                                       */
/*                                                                              */
/*                                                                              */
/********************************************************************************/

void optimization(X,x1,x2,fin_mat,rc,tot_eq,a1_b,domains)
MATRIX fin_mat, domains;
VECTOR X,a1_b;
IVECTOR x1,x2;
INDEX rc;
int tot_eq;
{
  extern double (*func4g)(double *);
  extern void (*func4boundarycheck)(double *);
  extern int *onboundary;
  extern int run_nochange_lim, run_notdone_add, run_notdone_lim;
  extern int converged_flag;
  extern double *parmvector;
  extern int usebfgsflag;

  static int firsttime = 0;

  static MATRIX
    new_genera,   /*Temporary storage for the new generation*/
    population,   /*Population of x2 variables*/
    print_pop,
    temp;

  static VECTOR
    probab,       /*Probability of agents to die or live*/
    cum_probab,   /*Cumulative probability of agents*/
    t_vec;

  static IVECTOR
    live,        /* live elements in population */
    parents;     /*Indexes of the x2_vari parents for crossover*/


  static int
    pop_size,     /*Population size*/
    MinMax,       /*Minimization or Maximization problem 0 or 1 */
    P,            /*Total number of agents chosen to reproduce*/
    P1,           /*Number of parents for each of the 8 operators*/
    P2,
    P3,
    P4,
    P5,
    P6,
    P7,
    P8,
    B,            /*Parameter for the 3rd operator - nonuniform mutation*/
    STEP;         /*Parameter for the 5th operator - simple arithmetical crossover*/

  static unsigned long
    generations;    /*Total number of Generations*/

  static double
    Q;                   /*Probability of the best agent*/

  unsigned long count_gener= 1; /*Counter to keep track of the number of generations*/
  unsigned long peak_cnt;

  int
      j1,
      j2,
      j3,
      j4,
      j5,
      j6,
      j7,
      j8,
      oper,
      ocnt,
      x2_vari = rc.c-2,
      first,                 /*Index of the parent to mutate*/
      first_live,            /*Index of the two parents for crossover parents*/
      second_live,
      first_die,             /*Index of the two parents for crossover death*/
      second_die,
      die_now,               /*index of agent to replace in current operation*/
      tot,                   /*total number of chosen parents not used*/
      p2use,
      i,
      j,
      k;

  double
    A,            /*Parameter for the 4th operator - whole arithmetical cross over*/
    Teval,        /*Evaluation of the best agent*/
    initial_val,  /*Initial value of the population */
    peak_val;
  int x,y;

  FLAG  same;
  double**new;

  double diff;
  int nochange_gen;
  double oldfitvalue;

  double grad[VARS], evalX[VARS];

  int gradient_tolerance;
  double gradtol, mgradtol;

  int evaliter, hflag, btest, ii, boundary;

  double bfgsfit, finalhessin[VARS][VARS], bfgsoutX[VARS+1], evalgtol,
  bfgs_portion, portion, work[VARS+1];

  double popwrk, popstat, popmean, popvar ;

  nochange_gen=0;
  portion = 0.1;
  gradtol = 0.0000001;
  mgradtol = -gradtol;

  if (firsttime == 0) {
    /*Reading from the file:
      the population size,
      total number of generations,
      total number of times to apply each of the operators,
      probability of the best agent,
      minimization or maximization problem,
      the parameters for the operators
    */
    fscanf(input,"%d %lU %d %d %d %d %d %d %d %d",
	   &pop_size,&generations,&P1,&P2,&P3,&P4,&P5,&P6,&P7,&P8);
    fscanf(input,"%lf ",&Q);
    fscanf(input,"%d ",&MinMax);
    fscanf(input,"%d ",&B);
    fscanf(input,"%d ",&STEP);
    fscanf(input,"%d ",&test_num);
    fclose(input);

    /* make sure P5, P7 are even numbers */
    i = P5 / 2;
    if (P5 != i*2) P5++;
    i = P7 / 2;
    if (P7 != i*2) P7++;

    /*P is the total number of parents needed for applying all the operators*/
    P = P1 + P2 + P3 + P4 + P5 + P6 + P7 + P8;
    if(P >= pop_size) pop_size = P + 1;

    /* set generations limit */
    run_notdone_lim = generations;

    fprintf(output,"\n\n");
    fprintf(output,"Test case number      : %d\n", test_num);
    fprintf(output,"Number of operators   : %d  %d  %d  %d  %d  %d  %d  %d\n",
	    P1, P2, P3, P4, P5, P6, P7, P8);
    fprintf(output,"Number of generations : %lu\n",generations);
    fprintf(output,"Population size       : %d\n", pop_size);
    fprintf(output,"Parameter B           : %d\n", B);
    fprintf(output,"Parameter Q           : %f\n", Q);

    /*Space allocation for all the vectors and matrices involved*/
    population = matrix(1,pop_size,0,x2_vari+1);
    print_pop  = matrix(1,pop_size,0,x2_vari+tot_eq+1);
    new_genera = matrix(1,pop_size,1,x2_vari+1);
    temp       = matrix(1,x2_vari>2?x2_vari:2,1,x2_vari);
    probab     = vector(1,pop_size);
    t_vec      = vector(1,x2_vari);
    cum_probab = vector(1,pop_size);
    live       = ivector(1,pop_size);
    parents    = ivector(1,x2_vari);

    /* The new initial value matrix:
       setting a new initial value for every individual */
    for(j=1; j<=pop_size; j++) {
      for(i=1; i<=x2_vari; i++) { 
	population[j][i] = frange_ran(domains[i][1], domains[i][3]); 
	population[j][x2_vari+1] = 1.0;
      }
    }
    /* optionally set the first individual to values from a file */
    setparms(population[1]);

    /*Evaluate each of the agents in the new population*/
    for(i=1; i<=pop_size; i++) {
      for(j=1; j<=x2_vari; j++)
	X[j] = population[i][j];
      population[i][0] = evaluate(X);
    }

    /*Sort the initial inidivduals based on their evaluation function*/
    sort(MinMax,population,pop_size);
  }
  firsttime++;

  /* set generations limit */
  generations = run_notdone_lim;

  peak_val = 0.0;
  peak_cnt = 0;

  switch(MinMax) {
  case 0:
    Teval = population[1][0];
    fprintf(output,"%7lu \t%18.8f\n",count_gener,population[1][0]); 
    peak_cnt = count_gener;
    peak_val = population[1][0];
    break;
  case 1:
    Teval = population[1][0];
    fprintf(output,"%7lu \t%18.8f\n",count_gener,population[1][0]);
    peak_cnt = count_gener;
    peak_val = population[1][0];
    break;
  }

  fprintf(output,"\nThe 2 best initial individuals are\n");
  for(i=1; i<3; i++) {
    print_vector(population[i],1,tot_eq+x2_vari);
    fprintf(output,"\nfitness = %f", population[i][0]);
    fprintf(output,"\n\n");
  }

  fprintf(output,"\nThe worst fit of the population is:%f\n", 
      population[pop_size][0]);
  fprintf(output,"\n\n");

  fprintf(output,"\n\nGeneration#\t    Solution Value\n");
  fprintf(output,"\n%7d \t%18.8f\n", 0, population[1][0]);

  /*Assigning probability of survival for each of the agent, with the*/
  /*probability provided by the user for the best agent*/
  assign_probab(probab,pop_size,Q); 

  /*Finding the cumulative probability of the agents*/
  find_cum_probab(cum_probab,probab,pop_size);

  /*Reproducing and evaluating for the total number of generations times*/
  do
    {
      fflush(stdout);
      /*Initializing the live vector*/
      for(j=1; j<=pop_size; j++)
        {
          live[j] = 0;
          for(i=0; i<=x2_vari + 1; i++)
            new_genera[j][i] = population[j][i];
        }

      /*Finding the agents that will die and the agents that will reproduce*/
      find_live(cum_probab,live,pop_size,P);
      /* set die_now counter to start replacements with the worst agent.
         use of die_now is okay if the entire population (except the previous
         best) is to be replaced in each generation */
      die_now = pop_size;

      j1=j2=j3=j4=j5=j6=j7=j8=0;

      /* main operator loop */
      while(j1+j2+j3+j4+j4+j5+j5+j6+j7+j7+j8 < P)
        {
          oper = irange_ran(1,8);
          switch (oper)
            {
              case 1:
                     /*Applying the first operator, uniform mutation*/
                    if (j1 < P1)
                      {
			/*Find one parent for mutation operator 1*/
			first_live  = find_parent(live,pop_size);
			/* check that agent to replace is in range */
			if (die_now < 2) {
			  printf("No agents to be replaced\n");
			  exit(1);
			}

			new_genera[die_now][x2_vari+1] = 1.0;
			for(i=1; i<=x2_vari; i++)
			  t_vec[i] = population[first_live][i];
			for (ocnt = irange_ran(1,x2_vari); ocnt>0; ocnt--)
			  oper1(t_vec,fin_mat,rc);
			for(i=1; i<=x2_vari; i++)
			  new_genera[die_now][i] = t_vec[i];
			die_now--;
			j1++;
		      }
                    break;
              case 2:
                    /*Applying the second operator, boundary mutation*/
                    if (j2 < P2)
                      {
                        /*Find one parent for mutation operator 2*/
                        first_live  = find_parent(live,pop_size);
			/* check that agent to replace is in range */
			if (die_now < 2) {
			  printf("No agents to be replaced\n");
			  exit(1);
			}

                        new_genera[die_now][x2_vari+1] = 2.0;
                        for(i=1; i<=x2_vari; i++)
                          t_vec[i] = population[first_live][i];
			oper2(t_vec,fin_mat,rc);
                        for(i=1; i<=x2_vari; i++)
                          new_genera[die_now][i] = t_vec[i];
			die_now--;
                        j2++;
                      }
                    break;
              case 3:
                    /*Applying the third operator, non-uniform mutation*/
                    if (j3 < P3)
                      {
                        /*Find one parent for mutation operator 3*/
                        first_live  = find_parent(live,pop_size);
			/* check that agent to replace is in range */
			if (die_now < 2) {
			  printf("No agents to be replaced\n");
			  exit(1);
			}

                        new_genera[die_now][x2_vari+1] = 3.0;
                        for(i=1; i<=x2_vari; i++)
                          t_vec[i] = population[first_live][i];
			for (ocnt = irange_ran(1,x2_vari); ocnt>0; ocnt--)
			  oper3(t_vec,fin_mat,rc,generations,count_gener,B);
                        for(i=1; i<=x2_vari; i++)
                          new_genera[die_now][i] = t_vec[i];
			die_now--;
                        j3++;
                      }
                    break;

              case 4:
                    /*Applying the fourth operator, whole arithmetical crossover*/
                    if (j4 < (int) P4)
                      {
                        /*Find max(2,x2_vari) parents for crossover operator 4*/
			p2use = x2_vari > 2 ? x2_vari : 2;
			for (i=1; i<p2use; i++) {
			  parents[i] = find_parent(live,pop_size);
			  live[parents[i]]++;  /* no decr. first p2use-1 parents */
			}
			parents[p2use] = find_parent(live,pop_size);
			/* check that agents to replace are in range */
			if (die_now < 2) {
			  printf("No agents to be replaced\n");
			  exit(1);
			}
			new_genera[die_now][x2_vari+1]  = 4.0;
			for(k=1; k<=p2use; k++)
			  for(i=1; i<=x2_vari; i++)
			    temp[k][i] = population[parents[k]][i];
			oper4(temp,p2use,x2_vari);
			for(i=1; i<=x2_vari; i++)
			  new_genera[die_now][i]  = temp[1][i];
			die_now--;
                        j4++;
                      }
                    break;
              case 5:
                    /*Applying the fifth operator, simple arithmetical crossover*/
                    if (j5 < (int) P5/2)
                      {
                        /*Find two distinct parents for crossover operator 5*/
                        first_live  = find_parent(live,pop_size);
                        second_live = find_parent(live,pop_size);
                        same = TRUE;
                        for(i=1; i<=x2_vari; i++)
                          if (population[first_live][i] != population[second_live][i])
                            same = FALSE;
			/* check that agents to replace are in range */
			if (die_now < 3) {
			  printf("Not enough agents to be replaced\n");
			  exit(1);
			}
			first_die   = die_now-- ;
			second_die  = die_now-- ;
			new_genera[first_die][x2_vari+1]  = 5.0;
			new_genera[second_die][x2_vari+1] = 5.0;
                        if (!same)
                          {
                            for(i=1; i<=x2_vari; i++)
                              {
                                temp[1][i] = population[first_live][i];
                                temp[2][i] = population[second_live][i];
                              }
                            oper5(temp[1],temp[2],STEP,rc,fin_mat);
                            for(i=1; i<=x2_vari; i++)
                              {
                                new_genera[first_die][i]  = temp[1][i];
                                new_genera[second_die][i] = temp[2][i];
                              }
                          }
			else {
			  /* copy agent chosen twice into two new indivs */
			  for(i=1; i<=x2_vari; i++) {
			    new_genera[first_die][i]  = 
			      population[first_live][i];
			    new_genera[second_die][i] = 
			      population[second_live][i];
			  }
			}
                        j5++;
                      }
                    break;
              case 6:
                    /*Applying the sixth operator, whole non-uniform mutation*/
                    if (j6 < P6)
                      {
                        /*Find one parent for mutation operator 6*/
                        first_live  = find_parent(live,pop_size);
			/* check that agent to replace is in range */
			if (die_now < 2) {
			  printf("No agents to be replaced\n");
			  exit(1);
			}

                        new_genera[die_now][x2_vari+1] = 6.0;
                        for(i=1; i<=x2_vari; i++)
                          t_vec[i] = population[first_live][i];
                        oper6(t_vec,fin_mat,rc,generations,count_gener,B);
                        for(i=1; i<=x2_vari; i++)
                          new_genera[die_now][i] = t_vec[i];
			die_now--;
                        j6++;
                      }
                    break;
              case 7:
                    /*Applying the seventh operator*/
                    if (j7 < (int) P7/2)
                      {
                        /*Find two distinct parents for operator 7*/
                        first_live  = find_parent(live,pop_size);
                        second_live = find_parent(live,pop_size);
                        same = TRUE;
                        for(i=1; i<=x2_vari; i++)
                          if (population[first_live][i] != population[second_live][i])
                            same = FALSE;
			/* check that agents to replace are in range */
			if (die_now < 3) {
			  printf("Not enough agents to be replaced\n");
			  exit(1);
			}
			first_die   = die_now-- ;
			second_die  = die_now-- ;
			new_genera[first_die][x2_vari+1]  = 7.0;
			new_genera[second_die][x2_vari+1] = 7.0;
                        if (!same) {
			  if (first_live < second_live)
			    /* first agent is better agent */
			    for(i=1; i<=x2_vari; i++) {
			      temp[2][i] = population[first_live][i];
			      temp[1][i] = population[second_live][i];
			    }
			  else
			    /* second agent is better agent */
			    for(i=1; i<=x2_vari; i++) {
			      temp[2][i] = population[second_live][i];
			      temp[1][i] = population[first_live][i];
			    }
			  oper7(temp[1],temp[2],rc,fin_mat);
			  for(i=1; i<=x2_vari; i++)
			    new_genera[first_die][i]  = temp[1][i];
			  if (first_live < second_live)
			    /* first agent is better agent */
			    for(i=1; i<=x2_vari; i++) {
			      temp[2][i] = population[first_live][i];
			      temp[1][i] = population[second_live][i];
			    }
			  else
			    /* second agent is better agent */
			    for(i=1; i<=x2_vari; i++) {
			      temp[2][i] = population[second_live][i];
			      temp[1][i] = population[first_live][i];
			    }
			  oper7(temp[1],temp[2],rc,fin_mat);
			  for(i=1; i<=x2_vari; i++)
			    new_genera[second_die][i]  = temp[1][i];
			}
			else {
			  /* copy agent chosen twice into two new indivs */
			  for(i=1; i<=x2_vari; i++) {
			    new_genera[first_die][i]  = 
			      population[first_live][i];
			    new_genera[second_die][i] = 
			      population[second_live][i];
			  }
			}
                        j7++;
                      }
              case 8:
                     /*Applying the eighth operator, BFGS*/
                    if (j8 < P8)
                      {
                        /*Find one parent for BFGS operator 1*/
                        first_live  = find_parent(live,pop_size);
			/* check that agent to replace is in range */
			if (die_now < 2) {
			  printf("No agents to be replaced\n");
			  exit(1);
			}

                        new_genera[die_now][x2_vari+1] = 8.0;
                        for(i=1; i<=x2_vari; i++)
                          t_vec[i] = population[first_live][i];
                        oper8(t_vec,fin_mat,rc,domains);
                        for(i=1; i<=x2_vari; i++)
                          new_genera[die_now][i] = t_vec[i];
			die_now--;
                        j8++;
                      }
                    break;
            }
        }


      /*Replace the population with the new generation */
      new = new_genera;
      new_genera = population;
      population = new;

      /*Evaluate each of the agents in the new population*/
      for(i=1; i<=pop_size; i++) {
	/* do not reevaluate if no changes were made */
	if (population[i][x2_vari+1] != 0) {
	  if (tot_eq != 0) {
	    for(j=1; j<=x2_vari + tot_eq; j++)
	      X[j] = 0.0;

	    find_X(fin_mat,population[i],X,x2_vari,tot_eq,x1,x2,a1_b);
	  }
	  else
	    for(j=1; j<=x2_vari; j++)
	      X[j] = population[i][j];

	  population[i][0] = evaluate(X);
	}
      }

      /*Sort the new population based on their evaluation function*/
      sort(MinMax,population,pop_size);

      /* print all generations */
      Teval = population[1][0];
      fprintf(output,"%7lu \t%18.8f\n",count_gener,population[1][0]); 
      peak_cnt = count_gener;
      peak_val = population[1][0];

      printbest(population[1]);

      if (count_gener == 1) {
	oldfitvalue=population[1][0];
      }

      diff=population[1][0]-oldfitvalue;
      if (MinMax==0) diff *= -1.0;

      if (diff < 0.000000001) {
	nochange_gen++;
      }
      else {
	nochange_gen=0;
      }

      if (nochange_gen > (run_nochange_lim-1) || count_gener >= run_notdone_lim) {
	gradient_tolerance=0;
	for (j=0; j<x2_vari; j++) 
	  {
	    evalX[j]=population[1][j+1];
	  }
	gradient(evalX, grad, x2_vari);
	func4boundarycheck(evalX-1);
	for (j=0;j<x2_vari;j++)
	  if (onboundary[j]==0) {
	    if (grad[j]>gradtol || grad[j]<mgradtol) gradient_tolerance=1;
	  }
	if (gradient_tolerance==1) nochange_gen=0;
      } /* end of gradient check for nochange_gen */

      if (nochange_gen > (run_nochange_lim-1)) {
	printf("NO REAL IMPROVEMENT IN %d GENERATIONS\n", run_nochange_lim);
	/* printf("REDUCING THE GENERATION LIMIT BY %d\n", NOCHANGE_SHRINK); */
	if (generations>NOCHANGE_SHRINK)
	  generations -= NOCHANGE_SHRINK;
	else {
	  generations = 0;
	  converged_flag = x2_vari;
	}
	nochange_gen=0;
      }

      /* increase the number of generations if fitness has been improving */
      if (count_gener == generations && nochange_gen < run_nochange_lim) {
	generations += run_notdone_add;
        printf("INCREASING GENERATION LIMIT BY %d\n", run_notdone_add);
      }

      if (count_gener >= run_notdone_lim) {
	if (gradient_tolerance==1) {
	  printf("GENERATION LIMIT OF %d REACHED\n", run_notdone_lim);
	  converged_flag = -1;
	}
	else
	  converged_flag = x2_vari;
        generations = 0;
      }

      if (usebfgsflag>0) {
	bfgs_portion=oldfitvalue*portion;
	if (count_gener<0) bfgs_portion=-1.0;

	printf("diff = %g, bfgs_portion = %g\n",diff, bfgs_portion);
	if (1 || diff<bfgs_portion) {
	  /* printf("In bfgs land\n"); */
	  if (0 && firsttime>1 && P2>0 && P8<5) {
	    P2--;
	    P8++;
	  } 
	  /* apply the bfgs to the best individual, if fit has changed */
	  if (1 || count_gener==1 || diff>0.0) {
	    for (i=1; i<=x2_vari; i++) {
	      bfgsoutX[i-1]=population[1][i];
	    }

	    boundary=0;
	    evalgtol=gradtol;
	    /* printf("start dfgsmin in evaluate.c\n");  fflush(stdout); */
	    dfgsmin(bfgsoutX, x2_vari, evalgtol, &evaliter, &bfgsfit,
		    finalhessin, domains);

	    for (ii=1; ii<=x2_vari; ii++) {
	      if ((bfgsoutX[ii-1]<domains[ii][1])
		  || (bfgsoutX[ii-1]>domains[ii][3])) {
		boundary=1;
	      }
	    }

	    if (boundary==0) {
	      if (bfgsfit<population[1][0]) {
		for(i=1;i<=x2_vari;i++) {
		  population[1][i]=bfgsoutX[i-1];
		}
		population[1][0]=bfgsfit;
	      }
	    }
	    else {
	      printf("OOB(%d \t%g): ", count_gener, bfgsfit);
	      printbest(bfgsoutX);
	    }
	    /* end of bfgs stuff */
	    printbest(population[1]);
	    peak_val = population[1][0];
	  }
	}
      }
      oldfitvalue=population[1][0];

      /* print current solution */
      find_X(fin_mat,population[1],X,x2_vari,tot_eq,x1,x2,a1_b);
      if (usebfgsflag<=0) {
	for(j = 1; j <= x2_vari+tot_eq; j++)
	  fprintf(output," X[%2d] =\t%14.9lf\n",j,X[j]);
      }
      else if (usebfgsflag>0) {
	for (j=0; j<x2_vari; j++) evalX[j]=X[j+1];
	estoptint(x2_vari, 9, 0, evalX, func4g);
	gradient(evalX, grad, x2_vari);
	for(j = 1; j <= x2_vari+tot_eq; j++)
	  fprintf(output," X[%2d] =\t%14.9lf \t %14.9lf\n",j,X[j], grad[j-1]);
      }
    }
  /*Increment iteration count and test whether all generations are done*/
  while (++count_gener <= generations);

  fprintf(output,
	  "\nBest solution was found at generation %lu (solution value = %.8f)\n",
	  peak_cnt,peak_val);
  fprintf(output,"\n\nBest solution found:\n\n");

  /* print best solution */
  find_X(fin_mat,population[1],X,x2_vari,tot_eq,x1,x2,a1_b);
  if (usebfgsflag<=0) {
    for(j = 1; j <= x2_vari+tot_eq; j++)
      fprintf(output," X[%2d] :\t%14.9lf\n",j,X[j]);
  }
  else if (usebfgsflag>0) {
    for (j=0; j<x2_vari; j++) evalX[j]=X[j+1];
    estoptint(x2_vari, 9, 1, evalX, func4g);
    gradient(evalX, grad, x2_vari);
    for(j = 1; j <= x2_vari+tot_eq; j++)
      fprintf(output," X[%2d] :\t%14.9lf \t %14.9lf\n",j,X[j], grad[j-1]);
  }

  /* copy parms to external vector for post-processing */
  for (j=0; j<x2_vari; j++) parmvector[j]=X[j+1];

  /* do cleanup computations at final value */
  cleanup(X);

  /* print final population */
/*
  for(i=1; i<=pop_size; i++)
    {
      if (tot_eq != 0)
        {
          for(j=1; j<=x2_vari + tot_eq; j++)
            X[j] = 0.0;

          find_X(fin_mat,population[i],X,x2_vari,tot_eq,x1,x2,a1_b);
          for(j=1; j<=x2_vari + tot_eq; j++)
            print_pop[i][j] = X[j];
        }
      else
        for(j=1; j<=x2_vari; j++)
          print_pop[i][j] = X[j] = population[i][j];

      print_pop[i][0] = evaluate(X);
      print_pop[i][x2_vari+tot_eq+1] = population[i][x2_vari+1];
    }

   fprintf(output,"\n\nThe Final Population:\n\n");
   fprintf(output,"  Solution Value\t          Agent variables: X[1] X[2] ...\n\n");
   print_population(output, pop_size,x2_vari+tot_eq+1,print_pop);
*/

}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   sort()                                       */
/*                                                                              */
/*           SYNOPSIS          :   void sort(MinMax, population,pop_size)       */
/*                                                                              */
/*           DESCRIPTION       :   This function sorts the population, in the   */
/*                                  ascending or the descending order of the    */
/*                                  evaluation function, depending on whether   */
/*                                  it is a maximization or a minimization      */
/*                                  function, respectively.                     */
/*                                                                              */
/*                                  As an alternative, the sortq function below */
/*                                  can be used, That sorting function uses     */
/*                                  the quicksort algorithm.                    */
/*                                                                              */
/*                                                                              */
/*           FUNCTIONS CALLED  :   swap()                                       */
/*                                                                              */
/*           CALLING FUNCITONS :   optimization()                               */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*          A               9/11/92       Tom Logan      Rewrote                */
/*                                                                              */
/********************************************************************************/
void sort(MinMax,population,pop_size)
int MinMax,       /*Tells whether it is a maximizaiton or a minimization function*/
    pop_size;     /*Population size*/
MATRIX population;/*Array of population*/
{
  int i,j,k;


  /*If MinMax is 0 sorts in the descending order, and*/
  /*if it is 1 sorts in the ascending order*/
  /*Sorted in ascending or descending order, based on*/
  /*the evaluated values of each of the agents*/
  switch(MinMax)
    {
    case 0 :
      for(i=1; i<=pop_size; i++)
        for(j=i+1; j<=pop_size; j++)
          if(population[i][0] > population[j][0])
            swap(&population[i],&population[j]);
      break;

    case 1 :
      for(i=1; i<=pop_size; i++)
        for(j=i+1; j<=pop_size; j++)
          if(population[i][0] < population[j][0])
            swap(&population[i],&population[j]);
      break;
    default:
      fprintf(output,"Incorrect data: Must be a 0 or 1");
      exit(1);
    }
}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   swap()                                       */
/*                                                                              */
/*           SYNOPSIS          :   void swap(x,y)                               */
/*                                                                              */
/*           DESCRIPTION       :   This function interchanges the values of     */
/*                                  x and y.                                    */
/*                                                                              */
/*           FUNCTIONS CALLED  :   None                                         */
/*                                                                              */
/*           CALLING FUNCITONS :   sort()                                       */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/



void swap(x,y)
double **x,**y;
{
  double *temp;

  temp = *x;
  *x = *y;
  *y = temp;
}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   find_parent()                                */
/*                                                                              */
/*           SYNOPSIS          :   int find_parent(live,pop_size)               */
/*                                                                              */
/*           DESCRIPTION       :   This function returns the index of the       */
/*                                  agent in the population, which is to be     */
/*                                  chosen for reproduction.                    */
/*                                                                              */
/*           FUNCTIONS CALLED  :   irange_ran()                                 */
/*                                                                              */
/*           CALLING FUNCITONS :   optimization()                               */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/

int find_parent(live,pop_size)
int pop_size; /*Population size*/
IVECTOR live; /*Vector containing the number of times each agent*/
              /*is going to reproduce*/
{
  int i,temp,t1,tot=0;

  /*Finding the total number of parents to reproduce*/
  for(i=1; i<=pop_size; i++)
    tot = tot + live[i];
  if(tot==0)
    {
      printf("No agents to select\n");
      exit(1);
    }

  /*Choosing one of them randomly*/
  temp = irange_ran(1,tot);

  tot = 0;
  i = 1;
  do{
    if(live[i]!=0)
      t1 = i;
    tot = tot + live[i++];
  }while(tot<temp);

  /*Decrementing the number of times the parent chosen is going to reproduce*/
  live[t1]--;
  return(t1);
}
/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   assign_probab()                              */
/*                                                                              */
/*           SYNOPSIS          :   void assign_probab(probab,pop_size,Q)        */
/*                                                                              */
/*           DESCRIPTION       :   This function assigns probability of survival*/
/*                                  to each of the agents determined by the     */
/*                                  value provided by the user for the          */
/*                                  probability of the best agnet.              */
/*                                                                              */
/*           FUNCTIONS CALLED  :   None                                         */
/*                                                                              */
/*           CALLING FUNCITONS :   optimization()                               */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/


void assign_probab(probab,pop_size,Q)
int pop_size;  /*Population size*/
double Q;       /*The probability of survival of the best agent*/
VECTOR probab; /*Array to contain the probability of survival*/
               /* of each of the agents*/
{
  int i;

  /* Q, Q(1-Q)^1, Q(1-Q)^2 ... Q(1-Q)^n */
  for(i=1; i<=pop_size; i++)
    probab[i] = Q * x_pow_y(1-Q,i-1);
}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   x_pow_y()                                    */
/*                                                                              */
/*           SYNOPSIS          :   double x_pow_y(x,y)                           */
/*                                                                              */
/*           DESCRIPTION       :   This function returns the value of x to the  */
/*                                  power of y.                                 */
/*                                                                              */
/*           FUNCTIONS CALLED  :   None                                         */
/*                                                                              */
/*           CALLING FUNCITONS :   assign_probab()                              */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/



double x_pow_y(x,y)
double x;
int y;
{
  int i;
  double tot = 1.0;

  for(i=0; i < y; i++)
    tot = tot * x;
  return(tot);
}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   find_cum__probab()                           */
/*                                                                              */
/*           SYNOPSIS          :   void find_cum__probab(cum_probab,probab,     */
/*                                                                     pop_size)*/
/*                                                                              */
/*           DESCRIPTION       :   This function finds the cumulative           */
/*                                  probability of each of the agents, from the */
/*                                  individual probability found earlier.       */
/*                                                                              */
/*           FUNCTIONS CALLED  :   None                                         */
/*                                                                              */
/*           CALLING FUNCITONS :   optimization()                               */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/



void find_cum_probab(cum_probab,probab,pop_size)
int pop_size;     /*Population size*/
VECTOR probab,    /*Individual probability of survival of each of the agent*/
       cum_probab;/*Cumulative probability of survival of each of the agent*/
{
  int i;

  cum_probab[1] = probab[1];

  for(i=2; i<=pop_size; i++)
    cum_probab[i] = cum_probab[i-1] + probab[i];

}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   find_live()                                  */
/*                                                                              */
/*           SYNOPSIS          :   void find_live(cum_probab,live,pop_size,P4+P5*/
/*                                                                              */
/*           DESCRIPTION       :   This function finds the agents from the      */
/*                                  population, who are going to live - those   */
/*                                  who are going to reproduce, which is done   */
/*                                  based on the cumulative probability of      */
/*                                  survival of each of the agents.             */
/*                                                                              */
/*           FUNCTIONS CALLED  :   frange_ran()                                 */
/*                                                                              */
/*           CALLING FUNCITONS :   optimization()                               */
/*                                                                              */
/*           AUTHOR            :   Tom Logan                                    */
/*                                                                              */
/*           DATE              :   9/15/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/



void find_live(cum_probab,live,pop_size,P)
VECTOR cum_probab; /*Cumulative probability*/
IVECTOR live;      /*Agents that are going to reproduce*/
int pop_size,      /*Population size*/
    P;             /*Total number of parents needed to reproduce*/
{
  double random;
  int count=0,/*Count of the number of agents chosen to live*/
      i;

  do
    {
      /*Choosing a random cumulative probability*/
      random = frange_ran(0.0,1.0);
      i=0;
      /*Finding the agent with the chosen cumulative probability*/
      do{
        i++;
        }while((random > cum_probab[i]) && (i< pop_size));

      /*Chosing the parent with that probability to reproduce*/
      if(count < P)
        {
          live[i]++;
          count++;
        }
    }while(count < P);
}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   find_die()                                   */
/*                                                                              */
/*           SYNOPSIS          :   void find_die(cum_probab,die,pop_size,P4+P5) */
/*                                                                              */
/*           DESCRIPTION       :   This function finds the agents from the      */
/*                                  population, who are going to die.            /
/*                                                                              */
/*           FUNCTIONS CALLED  :   frange_ran()                                 */
/*                                                                              */
/*           CALLING FUNCITONS :   optimization()                               */
/*                                                                              */
/*           AUTHOR            :   Tom Logan                                    */
/*                                                                              */
/*           DATE              :   9/15/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/



int find_die(cum_probab,die,pop_size)
VECTOR cum_probab; /*Cumulative probability*/
IVECTOR die;       /*Agents that are going to die*/
int pop_size;      /*Population size*/
{
  double random;
  int i;
  int done = FALSE;

  do
    {
      /*Choosing a random cumulative probability*/
      random = frange_ran(0.0,1.0);
      i=0;
      /*Finding the agent with the chosen cumulative probability*/
      do{
        i++;
        }
      while((random > cum_probab[i]) && (i< pop_size));

      /*Chosing the agent to die*/
      if ((die[pop_size+1-i] == 0) && (i < pop_size))
        done = TRUE;
    }
  while(!done);
  return(pop_size+1-i);
}

/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   find_X()                                     */
/*                                                                              */
/*           SYNOPSIS          :   void find_X(final,agen.X,x2_vari,tot_eq,x1,  */
/*                                                                           x2)*/
/*                                                                              */
/*           DESCRIPTION       :   This function finds the value of all the     */
/*                                  variables in the original problem, with     */
/*                                  the values of the remaining variables.      */
/*                                                                              */
/*           FUNCTIONS CALLED  :   none                                         */
/*                                                                              */
/*           CALLING FUNCITONS :   optimization()                               */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/



void find_X(final,agent,X,x2_vari,tot_eq,x1,x2,a1_b)
MATRIX final;
VECTOR agent,X,a1_b;
IVECTOR x1,x2;
int x2_vari,tot_eq;
{
  int i,j;

  for(j=1; j<=x2_vari; j++)
    X[x2[j]] = agent[j];

  for(j = 1; j<=tot_eq; j++)
    {
      X[x1[j]] = a1_b[j];
      for(i=1; i<=x2_vari; i++)
        X[x1[j]] = X[x1[j]] + agent[i] * final[j+x2_vari][i+1];
    }
}
