#include "gendop.h"

/* replacements for random number functions in GENOCOP's frange_ran.c */

#include "unif.h"
#include "urans.h"

static char *sccsversion = "%W%   %G%";

double newunif()
{
  static integer aux[TLPAUXSIZE], iseed = NEWUNIFSEED ;
  double wrkout;
  double wrk;

  /* get a random uniform double on (0,1] */
  ruxorv (&iseed, 1, &wrk, aux);
  wrkout = (double) wrk;
  return(wrkout);
}

double frange_ran(llim, ulim)
 double ulim;
 double llim;
{

  /*llim, ulim:  The upper and lower limits between which the random
      number is to be generated*/

  double num1, diff = ulim - llim;

  if (diff == 0)
    return(llim);
  else if(diff < 0.0001)
    return((flip() == TAIL) ? llim : ulim);
  do
    num1 = llim +  newunif()*(ulim-llim) ;
  while((num1<llim)||(num1>ulim));
  return(num1);
}

unsigned int randint ()
{
  static integer aux[TLPAUXSIZE], iseed = RANDINTSEED ;
  integer wrk;
  int num;

  /* get a signed 32-bit number from the TLP generator */
  tlpseq (&iseed, 1, &wrk, aux);
  /* truncate to 16 bits */
  num = wrk%65535;
  return (num);
}

unsigned int newrand ()
{
  return (randint());
}



/********************************************************************************/
/*                                                                              */
/*           FUNCTION NAME     :   initialize_x2()                              */
/*                                                                              */
/*           SYNOPSIS          :   FLAG initialize_x2(final,rc,x1,x2,x1_vari,   */
/*                                                                      x2_vari)*/
/*                                                                              */
/*           DESCRIPTION       :   This function generates random values for    */
/*                                  each of the x2 variables, in such a way that*/
/*                                  none of the constriants are violated; if no */
/*                                  such values could be found after a          */
/*                                  predetermined number of times, it returns   */
/*                                  FALSE.                                      */
/*                                                                              */
/*           FUNCTIONS CALLED  :   frange_ran(),                                */
/*                                 ivector(),                                   */
/*                                 matrix(),                                    */
/*                                 vector().                                    */
/*                                                                              */
/*           CALLING FUNCTIONS :   main()                                       */
/*                                                                              */
/*           AUTHOR            :   Swarnalatha Swaminathan                      */
/*                                                                              */
/*           DATE              :   1/17/92                                      */
/*                                                                              */
/*                                                                              */
/*           REV            DATE            BY           DESCRIPTION            */
/*           ---            ----            --           -----------            */
/*                                                                              */
/*                                                                              */
/********************************************************************************/


FLAG initialize_x2(final,rc,x1,x2,x1_vari,X,a1_b)
MATRIX final;        /*The final matrix*/
VECTOR X,a1_b;       /*Vector to hold the values of all the variables to be generated*/
IVECTOR x1,x2;       /*The p-variables eliminated and the remaining variables*/
INDEX rc;            /*Row and column of the final matrix*/
int x1_vari;         /*Number of variables eliminated*/
{
  int i,j,jasj, jasi,           /*counter variables*/
      x2_vari=rc.c-2,/*Total number of existing variables*/
      try=0;         /*Number of tries before getting a value for a variable*/
  FLAG _LOW,         /*Flags to check if the lower bounds are satisfied*/
       _HIGH;        /*Flags to check if the pper bounds are satisfied*/
  VECTOR temp;       /*Temporary vector to hold the random values of the variables*/
  MATRIX trymat;     /*Total after substitution of the generated variable value*/
  double llim,ulim,
        sum = 0;
  FLAG _CHECK = FALSE;


  temp = vector(1,x2_vari);
  trymat = matrix(1,rc.r-x2_vari,0,x2_vari);


  do
    {
      if(try < TRIES)
        /*Increment the number of tries*/
        ++try;
      for(i=1; i<=x2_vari; i++)
        /*For each constraints, generate a value for a variable between the bounds*/
        temp[i] = frange_ran(final[i][1],final[i][rc.c]);

      if(x2_vari != rc.r)
        {
          for(j=x2_vari+1; j<=rc.r; j++)
            {
              sum = 0;
              for(i=1; i<=x2_vari; i++)
                /*Evaluate the equation after a value is generated for a variable*/
                sum  =  sum + temp[i] * final[j][i+1];

              /*Check if the lower and upper bounds are satisfied*/
              _LOW = (sum >= final[j][1]) ? TRUE : FALSE;
              _HIGH = (sum <= final[j][rc.c]) ? TRUE : FALSE;

              if((!_LOW)||(!_HIGH))
                break;
            }
        }
      else
        _LOW = _HIGH = TRUE;
    }while((try<TRIES) && ((!_LOW)||(!_HIGH)));
  if(try >= TRIES)
    {
      printf("Please input initial values\n",x1_vari+x2_vari);
      for(i=1; i<=x1_vari+x2_vari; i++)
        {
          printf("\nX%d\t",i);
          scanf("%lf",&X[i]);
        }

      for(j=1; j<=rc.r; j++)
        {
          sum = 0;
          for(i=1; i<=x2_vari; i++)
            /*Evaluate the equation after a value is generated for a variable*/
            sum  =  sum + X[x2[i]] * final[j][i+1];

          /*Check if the lower and upper bounds are satisfied*/
          _LOW = (sum >= final[j][1]) ? TRUE : FALSE;
          _HIGH = (sum <= final[j][rc.c]) ? TRUE : FALSE;

          if((!_LOW)||(!_HIGH))
            {
              printf("The input values do not satisfy the constraints\n%d\n",j);
              exit(1);
            }
          else
            _CHECK = TRUE;
        }
    }
  else
    {
      _CHECK = TRUE;
      if(x1_vari != 0)
        {
          for(i=1; i<=x2_vari; i++)
            X[x2[i]] = temp[i];

          for(j=1; j<=x1_vari; j++)
            {
              X[x1[j]] = a1_b[j];
              for(i=1; i<=x2_vari; i++)
                X[x1[j]] = X[x1[j]] + temp[i] * final[j+x2_vari][i+1];
            }
        }
      else
        for(i=1; i<=x2_vari; i++)
            X[i] = temp[i];
    }

  free_vector(temp,1);
  free_matrix(trymat,1,rc.r-x2_vari,0);
  return(_CHECK);
}


