#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gendop.h"
#include "parms.h"
#include "objfunc.h"
#include "graddata.h"
#include "gradient.h"

static char *sccsversion = "%W%   %G%";

#define FNBUFSIZ 300

void (*func4genprint)(double *);
void (*func4setparms)(double *);
void (*func4setdomains)(double **);
double (*func4eval)(double *);
void (*func4cleanup)(double *);
double (*func4g)(double *);
double (*func4lhood)(double *);
int (*func4opg)(double *, double *);
void (*func4boundarycheck)(double *);

struct darray *thedata ;
int *onboundary;
double *parmvector;

int run_nochange_lim, run_notdone_add, run_notdone_lim;
int converged_flag;
int usebfgsflag;

void gendoph(void);

void main(int argc, char *argv[])
{
  extern void (*func4genprint)(double *);
  extern void (*func4setparms)(double *);
  extern void (*func4setdomains)(double **);
  extern double (*func4eval)(double *);
  extern void (*func4cleanup)(double *);
  extern double (*func4g)(double *);
  extern double (*func4lhood)(double *);
  extern int (*func4opg)(double *, double *);
  extern void (*func4boundarycheck)(double *);

  extern int *onboundary;
  extern int run_nochange_lim, run_notdone_add;

  int i,j,k, idx;
  FILE *fparms = NULL;

  int nparms, ndiffs;
  int nloops = 0;  /* number of times to repeat gendoph() */

  double *X, *invals, *dptr, **dummy;

  func4genprint = genprint;
  func4setparms = setoparms;
  func4setdomains = setdomains;
  func4eval = objfunc;
  func4cleanup = cleanupo;
  func4g = objfuncg;
  func4lhood = objfuncg;
  func4opg = objfuncopg;
  func4boundarycheck = boundarycheck;

  run_nochange_lim = NOCHANGE_LIM;
  run_notdone_add = NOTDONE_ADD;
  usebfgsflag=1;

  thedata = (struct darray *) calloc(1,sizeof(struct darray)) ;

  if (fparms != NULL) {
    readparms(fparms, MAX_NPARAMS, &thedata);
    nparms = thedata->nparms;
  }
  else {
    nparms = thedata->nparms = 1;
    thedata->parms = (double *) malloc(nparms*sizeof(double));
    thedata->parms[0] = 0.0;
  }
  parmvector = (double *) malloc(nparms*sizeof(double));
  for (i=0; i<nparms; i++) parmvector[i] = thedata->parms[i];

  X = (double *) malloc((nparms+1)*sizeof(double));
  invals = (double *) malloc(nparms*sizeof(double));

  ndiffs = 9;  /* order of differences for num grad optimal interval calcs */

  usebfgsflag=1;  /* =0, no bfgs;  =1, use bfgs */
  if (usebfgsflag>0) {
    /* set min, max, ranges so optimal interval calculations won't blow up */
    dptr = (double *) malloc(sizeof(double)*4);
    dummy = (double **) malloc((nparms+1)*sizeof(double *));
    for (i=0; i<(nparms+1); i++) dummy[i] = dptr;
    func4setdomains(dummy);

    /* allocate storage for vectors used in computing numerical gradients */
    epsacc = (double *) malloc(nparms*sizeof(double));
    optint = (double *) malloc(nparms*sizeof(double));
    ihessians = (double *) malloc(nparms*nparms*sizeof(double));

    /* estimate optimal intervals for the numerical gradients at starting values */
    estoptint(nparms, ndiffs, 1, parmvector, func4g);
    fflush(stdout);
  }

  onboundary = NULL;
  /* start GENOUD core */
  gendoph();

  printf("\nlog-likelihood = %lf\n\n", func4lhood(parmvector));

  for (j=0; j<nloops; j++) {
    /* reset starting values to get better intervals in numerical derivatives */
    for (i=0; i<nparms; i++) thedata->parms[i] = parmvector[i];

    usebfgsflag=1;

    estoptint(nparms, ndiffs, 1, parmvector, func4g);
    fflush(stdout);

    onboundary = NULL;
    gendoph();

    printf("\nlog-likelihood = %lf\n\n", func4lhood(parmvector));
  }

  dohessians(nparms, thedata->ncases, ndiffs, parmvector, func4lhood, func4opg);
  fflush(stdout);

  /* free storage allocated by calloc() for thedata structure */
  freethedata(&thedata);

} /* end main */

