#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "objfunc.h"
#include "parms.h"

static char *sccsversion = "%W%   %G%";

#define BAD_FIT_VALUE 99999999.0

/* number of parameters to estimate */
#define NPARAMS 28  /* ehopf4 */

/* read input data from a file */
void readthedata(FILE *infile, int ncases, struct darray **thedata) {
  int i, err;

  if (*thedata != NULL) {
    (*thedata)->y = (double *) calloc(ncases, sizeof(double)) ;

    for (i=0, err=0; i<ncases; i++) {
      err = fscanf(infile,"%lf", (*thedata)->y+i) ;
      if (err == EOF) break;
    }

    (*thedata)->ncases = i;
    printf("%d observations read\n", (*thedata)->ncases);
  }
}

/* read parameter starting values from a file */
void readparms(FILE *infile, int nparms, struct darray **thedata) {
  int i, err;

  if (*thedata != NULL) {
    (*thedata)->parms = (double *) calloc(nparms, sizeof(double)) ;

    for (i=0, err=0; i<nparms; i++) {
      err = fscanf(infile,"%lf",(*thedata)->parms+i) ;
      if (err == EOF) break;
    }

    (*thedata)->nparms = i;
    printf("%d parameters read\n", (*thedata)->nparms);
  }
}

/* read inverse hessian values from file.  optionally used in bfgs */
void readihess(FILE *infile, int nparms, struct darray **thedata) {
  int i, err, ihsize = nparms*nparms;

  if (*thedata != NULL) {
    (*thedata)->ihess = (double *) calloc(ihsize, sizeof(double)) ;

    for (i=0, err=0; i<ihsize; i++) {
      err = fscanf(infile,"%lf",(*thedata)->ihess+i) ;
      if (err == EOF) break;
    }

    printf("%d inverse hessian elements read\n", i);
  }
}

/* free allocated storage.  called at end of main() */
void freethedata(struct darray **thedata) {
}

/* genprint is called to print, optionally,
     the best vector of parameter values whenever the best fit value changes */
void genprint(double *X) {
}

/* initialize X vector of parameters;  called in optimization() */
void setoparms(double *X) {

  extern struct darray *thedata ;

  int i;

  if (thedata->parms != NULL)
    for (i=0; i<thedata->nparms; i++)
      X[i+1] = thedata->parms[i] ;
}

/* set particular domain limits as desired */
void setdomains(double **domains) {
}

/* do miscellaneous activities after optimization terminates */
void cleanupo(double *X) {
}

/* function to compute model log-likelihood or objective function.
   X[1] should be the first parameter. */
double objfunc(double *X) {
  extern struct darray *thedata ;

  double objective = 0.0;

  return objective ;
}

/* function to compute model log-likelihood or objective function.
   p[0] should be the first parameter.
   Used to compute numerical gradients */
double objfuncg(double *p) {
  return objfunc(p-1);
}

/* function to compute model log-likelihood or objective function,
   putting the value for each observation into *outvec.
   used to compute the outer product of the gradient at the final solution */
int objfuncopg(double *p, double *outvec) {
  extern struct darray *thedata ;

  int i, n = thedata->ncases;

  /* loop over data to compute objective function elements */
  for (i=0; i<n; i++) {
    outvec[i] = 0.0;
  }

  return n ;
}

void boundarycheck(double *X) {
  extern int *onboundary;

  int i;

  if (onboundary == NULL)
    onboundary = (int *) calloc(NPARAMS, sizeof(int));

  for (i=0; i<NPARAMS; i++)
    onboundary[i] = 0;

}
